#include <stdio.h> /* for DEBUGGING */
#include "mpi.h" /* for DEBUGGING */

/* This is example code of how to copy a large data structure to each
   slave.  This call to master_slave() would typically be done before
   the main call, in order to initialize data structures.  This is
   essentially a simulation of MPI_Bcast().  So, if you're running on
   top of MPI, you could alternatively call something like:
     MPI_Bcast ( data, 1+strlen(data), MPI_CHAR, 0, MPI_COMM_WORLD );
*/

#include "../mas-slave.h"

void *get_task() {
  static int first_time = 1;
  char *data = "VERY LARGE DATA STRUCTURE";
  if (first_time) {
    first_time = 0;
    return MSG(data, 1+strlen(data));
  }
  else return NOTASK;
}

void *do_task(task)
  void *task;
{ printf("in do task\n"); fflush(stdout);
  return (void *)0;
}

int get_task_result(result, task)
  void *result, *task;
{ printf("in get_task_result\n"); fflush(stdout);
  return UPDATE;
}

void update_environment(result, task)
  void *result, *task;
{
  int myrank;
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
printf("in update environment: "
	"myrank: %d; task size: %d; task: %s\n", myrank, task_size, task); fflush(stdout);
  printf("Node %d received: %s\n", myrank, (char *)task);
}

main(argc, argv)
  int argc;
  char *argv[];
{ char buffer[100];
  init_master_slave(2, &argc, &argv);
  /* Default receive buffer at least 1 KB, but this is nec. for larger msg's */
  declare_recv_buf(buffer, 100);
  master_slave(get_task, do_task, get_task_result, update_environment);
  exit(0);
}
